package be.iminds.ilabt.jfed.espec.parser;

import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.util.common.TextUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import org.jetbrains.annotations.Contract;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.stream.Collectors;

/**
 * ExperimentSpecificationParser
 */
public class ExperimentSpecificationParser {

    public static class ExperimentSpecificationParseException extends Exception {
        public ExperimentSpecificationParseException() { }
        public ExperimentSpecificationParseException(String s) { super(s); }
        public ExperimentSpecificationParseException(String s, Throwable throwable) { super(s, throwable); }
        public ExperimentSpecificationParseException(Throwable throwable) { super(throwable); }
        public ExperimentSpecificationParseException(String s, Throwable throwable, boolean b, boolean b1) { super(s, throwable, b, b1); }
    }

    private ObjectMapper mapper;
    public ExperimentSpecificationParser() {
        YAMLFactory yf = new YAMLFactory();
        this.mapper = new ObjectMapper(yf);
    }

    private int directCount;
    public ExperimentSpecification parse(String input) throws ExperimentSpecificationParseException {
        directCount = 0;

        Map root;
        try {
            root = mapper.readValue(input, Map.class);
        } catch (IOException e) {
            throw new ExperimentSpecificationParseException("Failed to parse ExperimentSpecification yaml", e);
        }
        if (root == null)
            throw new ExperimentSpecificationParseException("Failed to parse ExperimentSpecification yaml");

        ExperimentSpecificationBuilder res = new BasicExperimentSpecificationBuilder();

        checkAllowedKeys("root", root, "version", "rspec", "dir", "upload", "execute", "ansible");
        Object rawVersion = root.get("version");
        Object rawRspec = root.get("rspec");
        Object rawDir = root.get("dir");
        Object rawUpload = root.get("upload");
        Object rawExecute = root.get("execute");
        Object rawAnsible = root.get("ansible");

        if (rawVersion == null)
            throw new ExperimentSpecificationParseException("Experiment Specification must contain version");
        if (!(rawVersion instanceof String))
            throw new ExperimentSpecificationParseException("Experiment Specification version must be a String");

        try {
            res.setVersion((String) rawVersion);
        } catch (IllegalArgumentException e) {
            throw new ExperimentSpecificationParseException("Experiment Specification version must be a String " +
                    "of the form <base>-<flavor>", e);
        }

        res.setRspec(extractRspecSpec(rawRspec));
        res.setAnsible(extractAnsibleSpec(rawAnsible));
        res.setDirs(addDefaultDirsIfNeeded(extractDirSpecs(rawDir), res.getAnsible() != null));
        res.setUploads(extractUploadSpecs(rawUpload));
        res.setExecutes(extractExecuteSpecs(rawExecute));

        return res.create();
    }

    public static String makeValidDirPath(String dirPath) throws ExperimentSpecificationParseException {
        //path defaults to home dir  (for null as well as empty string)
        if (dirPath == null)
            return "~";
        if (dirPath.trim().isEmpty())
            return "~";

        if (dirPath.startsWith("/"))
            return dirPath;

        if (dirPath.startsWith("~")) {
            if (dirPath.length() == 1)
                return dirPath;
            if (dirPath.charAt(1) != '/')
                throw new ExperimentSpecificationParseException("Experiment Specification dir path \""+dirPath+"\" is not valid (starts with '"+dirPath.substring(0,2)+"')");
            return dirPath;
        }

        throw new ExperimentSpecificationParseException("Experiment Specification dir path \""+dirPath+"\" is not valid (does not start with / or ~)");
    }

    private List<DirSpec> addDefaultDirsIfNeeded(List<DirSpec> orig, boolean hasAnsible) {
        List<DirSpec> res = new ArrayList<>(orig);

        int scriptDirCount = 0;
        int uploadDirCount = 0;
        int ansibleDirCount = 0;

        DirSpec uploadDir = null;
        DirSpec scriptDir = null;
        for (DirSpec dir : orig) {
            if (dir.isUploadContent()) { uploadDirCount++; uploadDir = dir; }
            if (dir.isScriptContent()) { scriptDirCount++; scriptDir = dir; }
            if (dir.isAnsibleContent()) { ansibleDirCount++; }
        }

        //TODO check and fail if counts > 1

        if (uploadDirCount == 0) {
            //we can't check here if there is an upload dir needed on all nodes
            uploadDir = new DirSpec("~", ESpecConstants.DIRSPEC_CONTENT_UPLOAD,
                    "u=rwx", null);
            res.add(uploadDir);
        }
        assert uploadDir != null;
        if (scriptDirCount == 0) {
            //we can't check here if there is a script dir needed on all nodes
            scriptDir = new DirSpec(uploadDir.getPath(), ESpecConstants.DIRSPEC_CONTENT_SCRIPT,
                    "u=rwx", null);
            res.add(scriptDir);
        }
        assert scriptDir != null;
        if (ansibleDirCount == 0 && hasAnsible) {
            res.add(new DirSpec(ensureEndsWithSlash(scriptDir.getPath())+"ansible/",
                    ESpecConstants.DIRSPEC_CONTENT_ANSIBLE,
                    "u=rwx", null));
        }

        return res;
    }

    @Nonnull
    private List<DirSpec> extractDirSpecs(@Nullable Object rawDir) throws ExperimentSpecificationParseException {
        if (rawDir == null)
            return Collections.emptyList();

        if (rawDir instanceof String) {
            return Collections.singletonList(new DirSpec(makeValidDirPath((String) rawDir),
                    null, null, null));
        } else {
            if (rawDir instanceof Map) {
                return Collections.singletonList(mapToDirSpec((Map)rawDir, "dir"));
            } else {
                if (rawDir instanceof List) {
                    List list = (List) rawDir;

                    List<DirSpec> res = new ArrayList<>();
                    for (int i = 0; i < list.size(); i++) {
                        Object listObj = list.get(i);

                        if (listObj instanceof String) {
                            res.add(new DirSpec(makeValidDirPath((String) listObj),
                                    null, null, null));
                        } else {
                            if (listObj instanceof Map) {
                                res.add(mapToDirSpec((Map)listObj, "dir["+i+"]"));
                            } else {
                                throw new ExperimentSpecificationParseException(
                                        "Experiment Specification dir list must not have item of type " +
                                                "\"" + listObj.getClass().getName() + "\" (at index "+i+"))");
                            }
                        }
                    }
                    return res;
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification dir must not be of type " +
                            "\""+rawDir.getClass().getName()+"\")");
                }
            }
        }
    }

    @Nonnull
    private List<UploadSpec> extractUploadSpecs(@Nullable Object rawUpload) throws ExperimentSpecificationParseException {
        if (rawUpload == null)
            return Collections.emptyList();

        if (rawUpload instanceof String) {
            return Collections.singletonList(new UploadSpec(new FileSource(
                    FileSource.SourceType.BUNDLED, (String) rawUpload, (String) rawUpload),
                    null, null, null));
        } else {
            if (rawUpload instanceof Map) {
                return mapToUploadSpecs((Map)rawUpload, "upload");
            } else {
                if (rawUpload instanceof List) {
                    List list = (List) rawUpload;

                    List<UploadSpec> res = new ArrayList<>();
                    for (int i = 0; i < list.size(); i++) {
                        Object listObj = list.get(i);

                        if (listObj instanceof String) {
                            res.add(new UploadSpec(new FileSource(FileSource.SourceType.BUNDLED,
                                    (String) listObj, (String) listObj), null, null, null));
                        } else {
                            if (listObj instanceof Map) {
                                res.addAll(mapToUploadSpecs((Map)listObj, "upload["+i+"]"));
                            } else {
                                throw new ExperimentSpecificationParseException(
                                        "Experiment Specification upload list must not have item of type " +
                                                "\"" + listObj.getClass().getName() + "\" (at index "+i+"))");
                            }
                        }
                    }
                    return res;
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification upload must not be of " +
                            "type \""+rawUpload.getClass().getName()+"\")");
                }
            }
        }
    }

    @Nonnull
    private List<ExecuteSpec> extractExecuteSpecs(@Nullable Object rawExecute) throws ExperimentSpecificationParseException {
        if (rawExecute == null)
            return Collections.emptyList();

        if (rawExecute instanceof String) {
            return Collections.singletonList(new ExecuteSpec(new FileSource(FileSource.SourceType.BUNDLED,
                    (String) rawExecute, (String) rawExecute), null, null, null,
                    false, null, null));
        } else {
            if (rawExecute instanceof Map) {
                return Collections.singletonList(mapToExecuteSpec((Map)rawExecute, "execute"));
            } else {
                if (rawExecute instanceof List) {
                    List list = (List) rawExecute;

                    List<ExecuteSpec> res = new ArrayList<>();
                    for (int i = 0; i < list.size(); i++) {
                        Object listObj = list.get(i);

                        if (listObj instanceof String) {
                            res.add(new ExecuteSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) listObj,
                                    (String) listObj), null, null, null,
                                    false, null,null));
                        } else {
                            if (listObj instanceof Map) {
                                res.add(mapToExecuteSpec((Map)listObj, "execute["+i+"]"));
                            } else {
                                throw new ExperimentSpecificationParseException("Experiment Specification execute " +
                                        "list must not have item of type \"" + listObj.getClass().getName() + "\" " +
                                        "(at index "+i+"))");
                            }
                        }
                    }
                    return res;
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification execute must not be of " +
                            "type \""+rawExecute.getClass().getName()+"\")");
                }
            }
        }
    }

    @Nullable
    private AnsibleSpec extractAnsibleSpec(@Nullable Object rawAnsible) throws ExperimentSpecificationParseException {
        if (rawAnsible == null)
            return null;

        if (rawAnsible instanceof String) {
            return ansiblePlaybooksToAnsibleSpec(
                    Collections.singletonList(
                            stringToAnsiblePlaybookSpec((String) rawAnsible, "ansible"))
            );
        } else {
            if (rawAnsible instanceof Map) {
                return mapToAnsibleSpec((Map) rawAnsible);
            } else {
                if (rawAnsible instanceof List) {
                    List list = (List) rawAnsible;

                    List<AnsiblePlaybookSpec> res =
                            listToAnsiblePlaybookSpecList(list, "ansible");
                    return ansiblePlaybooksToAnsibleSpec(res);
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification execute must not be of " +
                            "type \""+rawAnsible.getClass().getName()+"\")");
                }
            }
        }
    }

    @Nonnull
    private DirSpec mapToDirSpec(@Nonnull Map in, @Nonnull String location) throws ExperimentSpecificationParseException {
        checkAllowedKeys(location, in, "path", "content", "permission", "nodes");

        String path = extractOptionalStringByKey(in, "path", location);
//        if (path == null)
//            throw new ExperimentSpecificationParseException("Experiment Specification "+location+" must have path)");
        String content = extractOptionalStringByKey(in, "content", location);
        String permissions = extractOptionalStringByKey(in, "permission", location);
        List<String> nodes = objectToNullableStringList(in.get("nodes"), location+".nodes");
        return new DirSpec(makeValidDirPath(path), content, permissions, nodes);
    }


    @Nonnull
    private List<UploadSpec> mapToUploadSpecs(@Nonnull Map in, @Nonnull String location) throws ExperimentSpecificationParseException {
        FileSource source = mapToFileSource(in, location, true);

        //special syntax for git: allow keys both as "git: <submap>", or on the same level as "git:" itself
        if (source.getType() == FileSource.SourceType.GIT || source.getType() == FileSource.SourceType.GITHUB)
            checkAllowedKeysWithFileSource(location, in, "path", "permission", "nodes", "url", "dir", "file", "branch", "username", "password", "privateKey");
        else
            checkAllowedKeysWithFileSource(location, in, "path", "permission", "nodes");

        String path = extractOptionalStringByKey(in, "path", location);
        String permissions = extractOptionalStringByKey(in, "permission", location);
        List<String> nodes = objectToNullableStringList(in.get("nodes"), location+".nodes");

        if (source.getType() == FileSource.SourceType.GENERATED) {
            if (source.isGeneratedKeyPair()) {
                if (path == null) {
                    //keypair has a default path
                    if (ESpecConstants.isKeyPair(source.getValue()) || ESpecConstants.isPrivateKey(source.getValue())) {
                        path = ESpecDefaults.KEYPAIR_PRIVATE_KEY_PATH;
                    }
                    if (ESpecConstants.isPublicKey(source.getValue())) {
                        path = ESpecDefaults.KEYPAIR_PUBLIC_KEY_PATH;
                    }
                }
                if (permissions == null) {
                    //keys have default permission for user read only
                    permissions = ESpecDefaults.KEYPAIR_PRIVATE_KEY_PERMISSIONS;
                }

                if (ESpecConstants.isKeyPair(source.getValue())) {
                    //split into 2 uploads: public and private key
                    FileSource privateSource = new FileSource(FileSource.SourceType.GENERATED,
                            ESpecConstants.KEYPAIR_PRIVATE, ESpecDefaults.KEYPAIR_PRIVATE_KEY_PATH);
                    FileSource publicSource = new FileSource(FileSource.SourceType.GENERATED,
                            ESpecConstants.KEYPAIR_PUBLIC, ESpecDefaults.KEYPAIR_PUBLIC_KEY_PATH);

                    return Arrays.asList(
                            new UploadSpec(privateSource, path, permissions, nodes),
                            new UploadSpec(publicSource, path + ".pub", permissions, nodes));
                }
            }
        }
        return Collections.singletonList(new UploadSpec(source, path, permissions, nodes));
    }


    @Nonnull
    private AnsibleSpec mapToAnsibleSpec(@Nonnull Map in) throws ExperimentSpecificationParseException {
        checkAllowedKeys("ansible", in, "host", "playbook", "galaxy", "group");

        Object rawHost = in.get("host");

        AnsibleHostSpec hostSpec;

        Object rawPlaybook = in.get("playbook");

        if (rawHost == null) {
            if (rawPlaybook == null)
                throw new ExperimentSpecificationParseException("Experiment Specification does not allow " +
                        "empty ansible.host and ansible.playbook at the same time");

            hostSpec = new AnsibleHostSpec(
                    AnsibleHostSpec.HostType.ADD,
                    ESpecDefaults.ANSIBLE_NODE_NAME,
                    ensureKeyPairInAnsibleUploads(null), addAnsibleInstallInAnsibleExecuteIfNeeded(null),
                    null, null);
        } else if (rawHost instanceof String) {
            AnsibleHostSpec.HostType ht = AnsibleHostSpec.HostType.valueOf(
                    extractMandatoryStringByKey(in, "type", "ansible.host")
                            .toUpperCase());

            hostSpec = new AnsibleHostSpec(
                    ht,
                    ESpecDefaults.ANSIBLE_NODE_NAME,
                    ensureKeyPairInAnsibleUploads(null), addAnsibleInstallInAnsibleExecuteIfNeeded(null),
                    null, null);
        } else if (rawHost instanceof Map) {
            Map hostMap = (Map) rawHost;
            checkAllowedKeys("ansible.host", hostMap,
                    "type", "name", "playbook-command", "galaxy-command", "upload", "execute");

            AnsibleHostSpec.HostType ht = AnsibleHostSpec.HostType.valueOf(
                    extractOptionalStringByKey(hostMap, "type", "ansible.host", "ADD")
                            .toUpperCase());

            String nodeName = extractOptionalStringByKey(hostMap, "name", "ansible.host");
            String galaxyCommand = extractOptionalStringByKey(hostMap, "playbook-command", "ansible.host");
            String playbookCommand = extractOptionalStringByKey(hostMap, "galaxy-command", "ansible.host");
            List<UploadSpec> uploadSpecs = hostMap.containsKey("upload") ?
                    extractUploadSpecs(hostMap.get("upload"))
                    : null;
            List<ExecuteSpec> executeSpecs = hostMap.containsKey("execute") ?
                    extractExecuteSpecs(hostMap.get("execute"))
                    : null;

            hostSpec = new AnsibleHostSpec(
                    ht,
                    nodeName == null ? ESpecDefaults.ANSIBLE_NODE_NAME : nodeName,
                    ensureKeyPairInAnsibleUploads(uploadSpecs),
                    addAnsibleInstallInAnsibleExecuteIfNeeded(executeSpecs),
                    galaxyCommand, playbookCommand);
        } else
            throw new ExperimentSpecificationParseException("Experiment Specification does not expect type " +
                    "\""+rawHost.getClass().getName()+"\" at ansible.host");

        List<AnsiblePlaybookSpec> ansiblePlaybookSpecs;

        if (rawPlaybook == null) {
            ansiblePlaybookSpecs = Collections.emptyList();
        } else if (rawPlaybook instanceof Map) {
            ansiblePlaybookSpecs = Collections.singletonList(
                    mapToAnsiblePlaybookSpec((Map)rawPlaybook, "ansible")
            );
        } else if (rawPlaybook instanceof List) {
            ansiblePlaybookSpecs =
                    listToAnsiblePlaybookSpecList((List)rawPlaybook, "ansible");
        } else if (rawPlaybook instanceof String) {
            ansiblePlaybookSpecs =
                    Collections.singletonList(
                            stringToAnsiblePlaybookSpec((String)rawPlaybook, "ansible"));
        } else
            throw new ExperimentSpecificationParseException("Experiment Specification does not allow " +
                    "ansible.playbook of type "+rawPlaybook.getClass().getName());

        Object rawGalaxy = in.get("galaxy");
        List<AnsibleGalaxySpec> ansibleGalaxySpecs;

        if (rawGalaxy == null) {
            ansibleGalaxySpecs = Collections.emptyList();
        } else if (rawGalaxy instanceof List) {
            ansibleGalaxySpecs =
                    listToAnsibleGalaxySpecList((List)rawGalaxy, "ansible");
        } else if (rawGalaxy instanceof String) {
            ansibleGalaxySpecs =
                    Collections.singletonList(
                            stringToAnsibleGalaxySpec((String)rawGalaxy, "ansible"));
        } else
            throw new ExperimentSpecificationParseException("Experiment Specification does not allow " +
                    "ansible.playbook of type "+rawGalaxy.getClass().getName());

        Map<String, List<String>> groups;
        Object rawGroups = in.get("group");
        if (rawGroups != null) {
            if (!(rawGroups instanceof Map))
                throw new ExperimentSpecificationParseException("ansible.group should not be of type "+
                        rawGroups.getClass().getName()+"");

            groups = checkGroups((Map)rawGroups);
        } else {
            groups = null;
        }

        return new AnsibleSpec(hostSpec,
                ansibleGalaxySpecs,
                ansiblePlaybookSpecs,
                groups
        );
    }

    /**
     * Checks if the groups Map contains the correct types
     * Also converts in some cases.
     *
     * @param groups
     * @return
     */
    private Map<String, List<String>> checkGroups(Map<?, ?> groups) throws ExperimentSpecificationParseException {
        Map<String, List<String>> res = new HashMap<>();

        for (Map.Entry e : groups.entrySet()) {
            Object key = e.getKey();
            Object val = e.getValue();
            if (!(key instanceof String))
                throw new ExperimentSpecificationParseException("ansible.group contains an entry with a " +
                        "non String key ("+key.getClass().getName()+")");
            if (!(val instanceof List || val instanceof String))
                throw new ExperimentSpecificationParseException("ansible.group contains an entry with a " +
                        "non String key ("+key.getClass().getName()+")");
            if (val instanceof List) {
                List nodeList = (List) val;
                for (Object nodeClientId : nodeList) {
                    if (!(nodeClientId instanceof String))
                        throw new ExperimentSpecificationParseException("ansible.group["+key+"] contains " +
                                "a list with a non String entry ("+key.getClass().getName()+")");
                }
                res.put((String) key, nodeList);
            } else {
                String nodeClientId = (String) val;
                res.put((String) key, Collections.singletonList(nodeClientId));
            }
        }

        return res;
    }

    private List<UploadSpec> ensureKeyPairInAnsibleUploads(@Nullable List<UploadSpec> orig) {
//        if (orig != null)
//            for (UploadSpec u : orig) {
//                if (u.getSource().isGeneratedKeyPair())
//                    return orig;
//            }
//        List<UploadSpec> used = orig == null ? new ArrayList<>() : new ArrayList<>(orig);
//        FileSource privateSource = new FileSource(FileSource.SourceType.GENERATED, "keypair.private", "id_rsa");
////        FileSource publicSource = new FileSource(FileSource.SourceType.GENERATED, "keypair.public", "id_rsa.pub");
//        used.add(new UploadSpec(privateSource, "~/.ssh/id_rsa", "600", null));
//        return used;

        //the above is not used: we'll upload all needed files in the playbook step (also if there is no playbook)
        return orig;
    }

    private List<ExecuteSpec> addAnsibleInstallInAnsibleExecuteIfNeeded(@Nullable List<ExecuteSpec> orig) {
        List<ExecuteSpec> used = new ArrayList<>();
        if (orig != null) {
            //I don't like the fake "auto" entry. Let's just assume that any existing extry in the list prevents
            // the auto install.
//            used = orig.stream()
//                    .filter(e -> e.getSource().getType() != FileSource.SourceType.BUNDLED
//                            || !e.getSource().getValue().trim().equalsIgnoreCase("auto"))
//                    .collect(Collectors.toCollection(ArrayList::new));
            used = new ArrayList<>(orig);
        } else {
            used = new ArrayList<>();
        }
        if (orig == null || orig.isEmpty() || orig.size() != used.size()) {
            //add internet access on vwall2 if needed is done in Experiment

            //add ansible auto install
            used.add(new ExecuteSpec(
                    new FileSource(
                            FileSource.SourceType.DOWNLOAD,
                            ESpecDefaults.ANSIBLE_INIT_SCRIPT_URL,
                            ESpecDefaults.ANSIBLE_INIT_SCRIPT_BASENAME),
                    null, null,
                    ESpecDefaults.ANSIBLE_INIT_SCRIPT_PERMISSIONS,
                    true,
                    null,
                    null));
        }
        return used;
    }


    @Nonnull
    private AnsibleSpec ansiblePlaybooksToAnsibleSpec(@Nonnull List<AnsiblePlaybookSpec> in) throws ExperimentSpecificationParseException {
        return new AnsibleSpec(new AnsibleHostSpec(
                AnsibleHostSpec.HostType.ADD,
                ESpecDefaults.ANSIBLE_NODE_NAME,
                ensureKeyPairInAnsibleUploads(null), addAnsibleInstallInAnsibleExecuteIfNeeded(null),
                null, null),
                null,
                in,
                null
        );
    }
    @Nonnull
    private List<AnsiblePlaybookSpec> listToAnsiblePlaybookSpecList(@Nonnull List in, @Nonnull String location) throws ExperimentSpecificationParseException {
        List<AnsiblePlaybookSpec> res = new ArrayList<>();
        for (int i = 0; i < in.size(); i++) {
            Object listObj = in.get(i);

            if (listObj instanceof String) {
                res.add(stringToAnsiblePlaybookSpec((String) listObj, location+"["+i+"]"));
            } else {
                if (listObj instanceof Map) {
                    res.add(mapToAnsiblePlaybookSpec((Map)listObj, location+"["+i+"]"));
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification execute list must not " +
                            "have item of type \"" + listObj.getClass().getName() + "\" ("+location+" at index "+i+"))");
                }
            }
        }
        return res;
    }
    @Nonnull
    private AnsiblePlaybookSpec stringToAnsiblePlaybookSpec(@Nonnull String in, @Nonnull String location) throws ExperimentSpecificationParseException {
        return new AnsiblePlaybookSpec(
                new FileSource(FileSource.SourceType.BUNDLED, in, in),
                null, null, ESpecDefaults.ANSIBLE_DEBUG, null
        );
    }
    @Nonnull
    private List<AnsibleGalaxySpec> listToAnsibleGalaxySpecList(@Nonnull List in, @Nonnull String location) throws ExperimentSpecificationParseException {
        List<AnsibleGalaxySpec> res = new ArrayList<>();
        for (int i = 0; i < in.size(); i++) {
            Object listObj = in.get(i);

            if (listObj instanceof String) {
                res.add(stringToAnsibleGalaxySpec((String) listObj, location+"["+i+"]"));
            } else {
                if (listObj instanceof Map) {
                    res.add(mapToAnsibleGalaxySpec((Map)listObj, location+"["+i+"]"));
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification execute list must not " +
                            "have item of type \"" + listObj.getClass().getName() + "\" ("+location+" at index "+i+"))");
                }
            }
        }
        return res;
    }
    @Nonnull
    private AnsibleGalaxySpec stringToAnsibleGalaxySpec(@Nonnull String in, @Nonnull String location) throws ExperimentSpecificationParseException {
        return new AnsibleGalaxySpec(
                new FileSource(FileSource.SourceType.BUNDLED, in, in),
                null, null, null
        );
    }
    @Nonnull
    private AnsiblePlaybookSpec mapToAnsiblePlaybookSpec(@Nonnull Map in, @Nonnull String location) throws ExperimentSpecificationParseException {
        checkAllowedKeysWithFileSource(location, in,
                "path", "permission", "debug", "log", "extra-vars", "extra-vars-from");

        FileSource source = mapToFileSource(in, location, false);
        String path = extractOptionalStringByKey(in, "path", location);
        checkRequirePathOrSource(source, path, location);
        String permissions = extractOptionalStringByKey(in, "permission", location);
        int debug = extractOptionalBoundedIntByKey(in, "debug", ESpecDefaults.ANSIBLE_DEBUG, 0, 3, location);
        String log = extractOptionalStringByKey(in, "log", location);

        Object extraVars = in.get("extra-vars");
        Object extraVarsFrom = in.get("extra-vars-from");

        if (extraVars != null && extraVarsFrom != null)
            throw new ExperimentSpecificationParseException("extra-vars and extra-vars-from may not be used together at "+location);

        if (extraVars == null && extraVarsFrom == null)
            return new AnsiblePlaybookSpec(source, path, permissions, debug, log);

        if (extraVars != null) {
            if (extraVars instanceof Map)
                return new AnsiblePlaybookSpec(source, path, permissions, debug, log,
                        (Map) extraVars);
            if (extraVars instanceof String) //direct in --extra-vars format
                return new AnsiblePlaybookSpec(source, path, permissions, debug, log,
                        (String) extraVars);

            throw new ExperimentSpecificationParseException("extra-vars should not be of " +
                    "type " + extraVars.getClass().getName() + " at " + location);
        }


        String extraVarsFromPath = extractOptionalStringByKey(in, "path", location);
        String extraVarsFromPermissions = extractOptionalStringByKey(in, "permission", location);
        if (extraVarsFrom instanceof Map)
            return new AnsiblePlaybookSpec(source, path, permissions, debug, log,
                    new AnsiblePlaybookSpec.ExtraVarsJsonFileSpec(
                            mapToFileSource((Map) extraVarsFrom, location, true),
                            extraVarsFromPath, extraVarsFromPermissions, null
                    )
            );
        if (extraVarsFrom instanceof String)
            return new AnsiblePlaybookSpec(source, path, permissions, debug, log,
                    new AnsiblePlaybookSpec.ExtraVarsJsonFileSpec(
                            new FileSource(FileSource.SourceType.BUNDLED,
                                    (String) extraVarsFrom, (String) extraVarsFrom),
                            extraVarsFromPath, extraVarsFromPermissions, null
                    )
            );

        throw new ExperimentSpecificationParseException("extra-vars-from should not be of " +
                "type "+extraVarsFrom.getClass().getName()+" at "+location);
    }

    private void checkRequirePathOrSource(@Nullable FileSource source, @Nullable String path, @Nonnull String location) throws ExperimentSpecificationParseException {
        if (source == null && path == null)
            throw new ExperimentSpecificationParseException("at "+location+" either a path or a source (or both) is required.");
    }

    @Nonnull
    private AnsibleGalaxySpec mapToAnsibleGalaxySpec(@Nonnull Map in, @Nonnull String location) throws ExperimentSpecificationParseException {
        checkAllowedKeysWithFileSource(location, in, "path", "permission", "log");

        FileSource source = mapToFileSource(in, location, false);
        String path = extractOptionalStringByKey(in, "path", location);
        checkRequirePathOrSource(source, path, location);
        String permissions = extractOptionalStringByKey(in, "permission", location);
        String log = extractOptionalStringByKey(in, "log", location);
        return new AnsibleGalaxySpec(source, path, permissions, log);
    }

    @Nonnull
    private ExecuteSpec mapToExecuteSpec(@Nonnull Map in, @Nonnull String location) throws ExperimentSpecificationParseException {
        checkAllowedKeysWithFileSource(location, in, "path", "permission", "pwd", "nodes", "sudo", "log");

        FileSource source = mapToFileSource(in, location, false);
        String path = extractOptionalStringByKey(in, "path", location);
        checkRequirePathOrSource(source, path, location);
        String pwd = extractOptionalStringByKey(in, "pwd", location);
        String permissions = extractOptionalStringByKey(in, "permission", location);
        String log = extractOptionalStringByKey(in, "log", location);
        boolean sudo = extractOptionalBooleanByKey(in, "sudo", false, location);
        List<String> nodes = objectToNullableStringList(in.get("nodes"), location+".nodes");
        return new ExecuteSpec(source, path, pwd, permissions, sudo, log, nodes);
    }

    @Nullable
    @Contract("null,_ -> null; !null,_ -> !null")
    private List<String> objectToNullableStringList(@Nullable Object in, @Nonnull String location) throws ExperimentSpecificationParseException {
        if (in == null)
            return null;

        if (in instanceof String) {
            return Collections.singletonList((String) in);
        } else {
            if (in instanceof List) {
                List list = (List) in;
                List<String> res = new ArrayList<>(list.size());
                for (int i = 0; i < list.size(); i++) {
                    Object listItemObj = list.get(i);
                    if (listItemObj instanceof String) {
                        res.add((String) listItemObj);
                    } else
                        throw new ExperimentSpecificationParseException("Experiment Specification expect type String " +
                                "instead of \""+in.getClass().getName()+"\" at "+location+"["+i+"]");
                }
                return res;
            } else {
                throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                        "type \""+in.getClass().getName()+"\" at "+location);
            }
        }
    }

    @Nullable
    @Contract("_,_,_,!null -> !null")
    private String extractOptionalStringByKey(@Nonnull Map m, @Nonnull String key,
                                              @Nonnull String location,
                                              @Nullable String defaultVal) throws ExperimentSpecificationParseException {
        String res = extractOptionalStringByKey(m, key, location);
        if (res == null)
            return defaultVal;
        else
            return res;
    }

    @Nullable
    private String extractOptionalStringByKey(@Nonnull Map m, @Nonnull String key, @Nonnull String location) throws ExperimentSpecificationParseException {
        Object in = m.get(key);
        if (in == null)
            return null;

        if (in instanceof String) {
            return (String) in;
        } else {
            throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                    "type \""+in.getClass().getName()+"\" at "+location+"."+key);
        }
    }

    private boolean extractOptionalBooleanByKey(@Nonnull Map m, @Nonnull String key, boolean defaultVal, @Nonnull String location) throws ExperimentSpecificationParseException {
        Object in = m.get(key);
        if (in == null)
            return defaultVal;

        if (in instanceof Boolean) {
            return (Boolean) in;
        } else if (in instanceof String) {
            Boolean parsed = TextUtil.objectToBoolean(in);
            if (parsed != null)
                return parsed;
            else
                throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                        "type \""+in.getClass().getName()+"\" at "+location+"."+key);
        } else {
            throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                    "type \""+in.getClass().getName()+"\" at "+location+"."+key);
        }
    }

    private int extractOptionalBoundedIntByKey(@Nonnull Map m, @Nonnull String key, int defaultVal, int minVal, int maxVal, @Nonnull String location) throws ExperimentSpecificationParseException {
        Object in = m.get(key);
        if (in == null)
            return defaultVal;

        int res;
        if (in instanceof Number) {
            res = ((Number) in).intValue();
        } else if (in instanceof String) {
            try {
                res = Integer.parseInt((String) in);
            } catch (NumberFormatException e) {
                throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                        "String \"" + in + "\" at " + location + "." + key);
            }
        } else {
            throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                    "type \""+in.getClass().getName()+"\" at "+location+"."+key);
        }

        if (res < minVal || res > maxVal)
            throw new ExperimentSpecificationParseException("Experiment Specification got  " + res+" " +
                    "but expected a value between "+minVal+" and "+maxVal+" at " + location + "."+key);

        return res;
    }

    @Nonnull
    private String extractMandatoryStringByKey(@Nonnull Map m, @Nonnull String key, @Nonnull String location) throws ExperimentSpecificationParseException {
        Object in = m.get(key);
        if (in == null)
            throw new ExperimentSpecificationParseException("Experiment Specification expected key \""+key+"\" " +
                    "at "+location);

        if (in instanceof String) {
            return (String) in;
        } else {
            throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                    "type \""+in.getClass().getName()+"\" at "+location+"."+key);
        }
    }

    private long extractMandatoryLongByKey(@Nonnull Map m, @Nonnull String key, @Nonnull String location) throws ExperimentSpecificationParseException {
        Object in = m.get(key);
        if (in == null)
            throw new ExperimentSpecificationParseException(
                    "Experiment Specification expected key \""+key+"\" at "+location);

        if (in instanceof Integer) {
            return ((Integer) in).longValue();
        } else {
            if (in instanceof Long) {
                return (Long) in;
            } else
                throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                        "type \""+in.getClass().getName()+"\" at "+location+"."+key);
        }
    }
    private int extractMandatoryIntByKey(@Nonnull Map m, @Nonnull String key, @Nonnull String location) throws ExperimentSpecificationParseException {
        Object in = m.get(key);
        if (in == null)
            throw new ExperimentSpecificationParseException(
                    "Experiment Specification expected key \""+key+"\" at "+location);

        if (in instanceof Integer) {
            return (Integer) in;
        } else {
            if (in instanceof Long) {
                return ((Long) in).intValue();
            } else
                throw new ExperimentSpecificationParseException("Experiment Specification did not expect " +
                        "type \""+in.getClass().getName()+"\" at "+location+"."+key);
        }
    }

    @Nullable
    @Contract("null,_ -> null; !null,_ -> !null")
    private String objectToString(@Nullable Object in, @Nonnull String location) throws ExperimentSpecificationParseException {
        if (in == null)
            return null;

        if (in instanceof String) {
            return (String) in;
        } else {
            throw new ExperimentSpecificationParseException(
                    "Experiment Specification did not expect type \""+in.getClass().getName()+"\" at "+location);
        }
    }

    @Nonnull
    private RspecSpec extractRspecSpec(@Nullable Object rawRspec) throws ExperimentSpecificationParseException {
        if (rawRspec == null)
            throw new ExperimentSpecificationParseException("Experiment Specification must contain rspec");

        if (rawRspec instanceof String) {
            return new RspecSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) rawRspec, (String) rawRspec));
        } else {
            if (rawRspec instanceof Map) {
                return mapToRspecSpec((Map)rawRspec, "rspec");
            } else {
                if (rawRspec instanceof List) {
                    List rspecList = (List) rawRspec;

                    if (rspecList.size() != 1)
                        throw new ExperimentSpecificationParseException("Experiment Specification rspec must not " +
                                "be list of size 1, not of size "+rspecList.size()+"");

                    Object listObj = rspecList.get(0);
//                    for (int i = 0; i < rspecList.size(); i++) {
//                        Object listObj = rspecList.get(i);

                    if (listObj instanceof String) {
                        return new RspecSpec(new FileSource(FileSource.SourceType.BUNDLED, (String) listObj, (String) listObj));
                    } else {
                        if (listObj instanceof Map) {
                            return mapToRspecSpec((Map)listObj, "rspec");
                        } else {
                            throw new ExperimentSpecificationParseException("Experiment Specification rspec list " +
                                    "must not have item of type \""+listObj.getClass().getName()+"\")");
                        }
                    }
                } else {
                    throw new ExperimentSpecificationParseException("Experiment Specification rspec must not be of " +
                            "type \""+rawRspec.getClass().getName()+"\")");
                }
            }
        }
    }


    @Nonnull
    private RspecSpec mapToRspecSpec(@Nonnull Map in, String location) throws ExperimentSpecificationParseException {
        if (in.containsKey("path"))
            throw new ExperimentSpecificationParseException("Experiment Specification rspec must not " +
                    "contain \"path\"");
        //only file sources are allowed for "rspec"
        checkAllowedKeysWithFileSource("rspec", in);

        return new RspecSpec(mapToFileSource(in, "rspec", true));
    }

    @Nullable
    @Contract("_,_,true->!null")
    private FileSource mapToFileSource(@Nonnull Map in, String location, boolean required) throws ExperimentSpecificationParseException {
        FileSource res = null;
        for (FileSource.SourceType sourceType : FileSource.SourceType.values()) {
            String key = sourceType.name().toLowerCase();
            Object val = in.get(key);
            if (val != null) {
                if (res != null)
                    throw new ExperimentSpecificationParseException(
                            "There are multiple file sources at "+location+" (at least "+
                                    key+" and "+res.getType().name().toLowerCase()+")");

                if (val instanceof String) {
                    if (sourceType == FileSource.SourceType.GENERATED) {
                        if (ESpecConstants.isKeyPair((String) val)) {
                            res = new FileSource(sourceType, (String) val, ESpecDefaults.KEYPAIR_PRIVATE_KEY_BASENAME);
                            continue;
                        }
                        if (ESpecConstants.isRandom((String) val)) {
                            throw new RuntimeException("generated: random requires an object as argument with fields 'method', 'format' and 'length'");
                        }
                        throw new ExperimentSpecificationParseException("Illegal GENERATED entry: " + key + ": " + val+"  at " + location);
                    }
                    if (sourceType == FileSource.SourceType.META) {
                        if (!ESpecConstants.isMetaValue((String) val)) {
                            throw new ExperimentSpecificationParseException("Illegal META entry: " + key + ": " + val+"  at " + location);
                        }
                    }
                    if (sourceType == FileSource.SourceType.GIT || sourceType == FileSource.SourceType.GITHUB) {
                        String gitUrl = (String) val;
                        res = extractGitFileSource(gitUrl, in, key, location);
                        continue;
                    }
                    res = new FileSource(sourceType, (String) val,
                            getFileSourceBasename(sourceType, (String) val, location, in));
                    continue;
                } else if (val instanceof Map) {
                    if (sourceType == FileSource.SourceType.GENERATED) {
                        Map subMap = (Map) val;
                        String method = extractMandatoryStringByKey(subMap, "method", location+"."+key);
                        if (ESpecConstants.isKeyPair(method)) {
                            res = new FileSource(sourceType, method, ESpecDefaults.KEYPAIR_PRIVATE_KEY_BASENAME);
                            continue;
                        }
                        if (ESpecConstants.isRandom(method)) {
                            String format = extractMandatoryStringByKey(subMap, "format", location+"."+key);
                            String basename =
                                    ESpecConstants.isRandomFormat(format, ESpecConstants.RANDOM_FORMAT_BINARY) ?
                                            ESpecDefaults.RANDOM_BINARY_BASENAME :
                                            ESpecDefaults.RANDOM_TXT_BASENAME;
                            res = new GeneratedRandomFileSource(
                                    sourceType, method, basename,
                                    format,
                                    extractMandatoryIntByKey(subMap, "length", location+"."+key));
                            continue;
                        }
                        if (ESpecConstants.isGeneratedRSpec(method)) {
                            String basename = "generated_experiment.rspec";

                            boolean multiAm;
                            Object amsObj = subMap.get("ams");
                            List<String> ams;
                            if (amsObj != null) {
                                if (amsObj instanceof String) {
                                    String amsStr = (String)amsObj;
                                    if (amsStr.trim().isEmpty()) {
                                        throw new ExperimentSpecificationParseException("Experiment Specification expected " +
                                                "key \"ams\" to be a string (with space seperated entries), " +
                                                "but it was an empty string at " + location + "." + key);
                                    }
                                    ams = Arrays.asList(amsStr.split(" "));
                                } else {
                                    if (!(amsObj instanceof List)) {
                                        throw new ExperimentSpecificationParseException("Experiment Specification expected " +
                                                "key \"ams\" to be a list of strings at " + location + "." + key);
                                    }
                                    ams = new ArrayList<>();
                                    List amsL = (List) amsObj;
                                    for (Object o : amsL) {
                                        if (o == null) continue;
                                        if (!(o instanceof String)) {
                                            throw new ExperimentSpecificationParseException("Experiment Specification found " +
                                                    "key \"ams\" which is a list containing a \"" + o.getClass().getName() + "\" " +
                                                    "at " + location + "." + key + "  (expected only strings in list)");
                                        }
                                        ams.add((String) o);
                                    }
                                }
                                multiAm = true;
                            } else {
                                multiAm = false;
                                ams = null;
                            }

                            Object resourceClassesObj = subMap.get("resourceClasses");
                            List<String> resourceClasses;
                            if (resourceClassesObj != null) {
                                if (!multiAm) {
                                    throw new ExperimentSpecificationParseException(
                                            "Experiment Specification found key \"resourceClasses\" " +
                                                    "without key \"ams\" specified " +
                                                    "at " + location + "." + key + "");
                                }
                                if (resourceClassesObj instanceof String) {
                                    String resourceClassesStr = (String)resourceClassesObj;
                                    if (resourceClassesStr.trim().isEmpty()) {
                                        throw new ExperimentSpecificationParseException("Experiment Specification expected " +
                                                "key \"resourceClasses\" to be a string (with space seperated entries), " +
                                                "but it was an empty string at " + location + "." + key);
                                    }
                                    resourceClasses = Arrays.asList(resourceClassesStr.split(" "));
                                } else {
                                    if (!(resourceClassesObj instanceof List)) {
                                        throw new ExperimentSpecificationParseException("Experiment Specification expected " +
                                                "key \"resourceClasses\" to be a list of strings at " + location + "." + key);
                                    }
                                    resourceClasses = new ArrayList<>();
                                    List resourceClassesL = (List) resourceClassesObj;
                                    for (Object o : resourceClassesL) {
                                        if (o == null) continue;
                                        if (!(o instanceof String)) {
                                            throw new ExperimentSpecificationParseException("Experiment Specification found " +
                                                    "key \"resourceClasses\" which is a list containing a \"" + o.getClass().getName() + "\" " +
                                                    "at " + location + "." + key + "  (expected only strings in list)");
                                        }
                                        resourceClasses.add((String) o);
                                    }
                                }
                            } else {
                                resourceClasses = null;
                            }

                            String nodeNamePrefix = extractOptionalStringByKey(subMap, "prefix", location+"."+key);
                            Object nodesObj = subMap.get("nodes");
                            List<String> nodeNames = null;
                            Integer nodeCount = null;
                            if (nodesObj != null) {
                                if (nodesObj instanceof Integer) {
                                    nodeCount = (Integer) nodesObj;
                                } else if (nodesObj instanceof Long) {
                                    nodeCount = ((Long) nodesObj).intValue();
                                } else if (nodesObj instanceof String) {
                                    nodeNames = Collections.singletonList((String) nodesObj);
                                } else if (nodesObj instanceof List) {
                                    List tmp = (List) nodesObj;
                                    nodeNames = new ArrayList<>(tmp.size());
                                    for (Object o : tmp) {
                                        if (o == null) continue;
                                        if (!(o instanceof String)) {
                                            throw new ExperimentSpecificationParseException("Experiment Specification found " +
                                                    "key \"nodes\" which is a list containing a \"" + o.getClass().getName() + "\" " +
                                                    "at " + location + "." + key + "  (expected only strings in list)");
                                        }
                                        nodeNames.add((String) o);
                                    }
                                } else {
                                    throw new ExperimentSpecificationParseException("Experiment Specification found " +
                                            "key \"nodes\" with unexpected type \"" + nodesObj.getClass().getName() + "\" " +
                                            "at " + location + "." + key + "  (expected integer, string or list of strings)");
                                }
                                if (nodeNames == null && multiAm && ams.size() != nodeCount) {
                                    throw new ExperimentSpecificationParseException("Experiment Specification found " +
                                            "key \"nodes\" which is the number " + nodeCount + " but this does not match the " +
                                            ams.size() + " ams specified " +
                                            "at " + location + "." + key + "");
                                }
                                if (nodeNames != null && multiAm && ams.size() != nodeNames.size()) {
                                    throw new ExperimentSpecificationParseException("Experiment Specification found " +
                                            "key \"nodes\" which contains " + nodeNames.size() + " node names but this does not match the " +
                                            ams.size() + " ams specified " +
                                            "at " + location + "." + key + "");
                                }
                            } else {
                                if (!multiAm)
                                    throw new ExperimentSpecificationParseException("Experiment Specification expected " +
                                            "key \"nodes\" at " + location + "." + key);
                            }

                            if (multiAm) {
                                //bad idea to make this large, as nodes are interlinked
                                if (ams.size() <= 1 || ams.size() > 10)
                                    throw new ExperimentSpecificationParseException("Experiment Specification " +
                                            "at "+location+"."+key+ ".\"ams\" has in invalid AM count (min 2, max 9) " +
                                            "(="+ams.size()+")");
                                if (resourceClasses != null && resourceClasses.size() != ams.size()) {
                                    throw new ExperimentSpecificationParseException("Experiment Specification " +
                                            "at "+location+"."+key+ ".\"ams\" has mismatch: AM count " +
                                            "(="+ams.size()+") != resourceClasses count (="+resourceClasses.size()+")." +
                                            "They need to match, or resourceClasses needs to be omitted.");
                                }

                                res = new GeneratedStitchingTestRSpecFileSource(
                                        sourceType, method, basename,
                                        ams, resourceClasses, nodeNamePrefix, nodeNames);
                            } else {
                                assert nodesObj != null;

                                if (nodeCount == null) {
                                    assert nodeNames != null;
                                    nodeCount = nodeNames.size();
                                }

                                if (nodeCount <= 0 || nodeCount > 1000)
                                    throw new ExperimentSpecificationParseException("Experiment Specification " +
                                            "at "+location+"."+key+ ".\"nodes\" has in invalid node count " +
                                            "(="+nodeCount+")");

                                String am = extractMandatoryStringByKey(subMap, "am", location + "." + key);
                                String resourceClass = extractOptionalStringByKey(subMap, "icon", location + "." + key);
                                String diskImage = extractOptionalStringByKey(subMap, "diskImage", location + "." + key);

                                Boolean link = extractOptionalBooleanByKey(subMap, "link", false, location + "." + key);
                                String linkSubnet = extractOptionalStringByKey(subMap, "linkSubnet", location + "." + key);
                                Integer linkStartIp = extractOptionalBoundedIntByKey(subMap, "linkStartIp", 1, 1, 254, location + "." + key);

                                res = new GeneratedRSpecFileSource(
                                        sourceType, method, basename,
                                        am, resourceClass,
                                        nodeCount, nodeNamePrefix, nodeNames,
                                        diskImage, link, linkSubnet, linkStartIp);
                            }
                            continue;
                        }
                        throw new ExperimentSpecificationParseException("Illegal GENERATED entry: " + key + ": val of type object = " + val);
                    }
                    if (sourceType == FileSource.SourceType.GIT || sourceType == FileSource.SourceType.GITHUB) {
                        res = extractGitFileSource(null, (Map) val, key, location);
                        continue;
                    }
                    throw new ExperimentSpecificationParseException("Illegal non-GENERATED entry: " + key + ": val of type object = " + val);
                } else {
                    throw new ExperimentSpecificationParseException("Illegal entry: " + key + ": " + val.getClass().getName());
                }
            }
        }

        if (res != null)
            return res;

        if (required)
            throw new ExperimentSpecificationParseException("No \"source\" (bundled, direct, ...) found in "+location+" " +
                    "(did find: "+in.keySet().stream().collect(Collectors.joining(","))+")");
        else
            return null;
    }

    @Nonnull
    private GitFileSource extractGitFileSource(@Nullable String defaultUrl,
                                               @Nonnull Map srcMap,
                                               @Nonnull String key,
                                               @Nonnull String location)
            throws ExperimentSpecificationParseException {
        boolean mandatoryUrl = defaultUrl == null;
        String gitUrl = mandatoryUrl ?
                extractMandatoryStringByKey(srcMap, "url", location+"."+key):
                extractOptionalStringByKey(srcMap, "url", location+"."+key);
        if (gitUrl == null)
            gitUrl = defaultUrl;
        String dir = extractOptionalStringByKey(srcMap,
                "dir", location+"."+key);
        String branch = extractOptionalStringByKey(srcMap,
                "branch", location+"."+key);
        String username = extractOptionalStringByKey(srcMap,
                "username", location+"."+key);
        String password = extractOptionalStringByKey(srcMap,
                "password", location+"."+key);
        String privateKey = extractOptionalStringByKey(srcMap,
                "privateKey", location+"."+key);
        String filename = extractOptionalStringByKey(srcMap,
                "file", location+"."+key);
        String basename = filename == null ? "git-repo" :
                filename.replaceAll(".*/", "");
        if (dir != null && filename != null) {
            filename = dir + (dir.endsWith("/") ? "" : "/") + filename;
            dir = null;
        }
        if ((username != null) != (password != null)) {
            throw new ExperimentSpecificationParseException("Experiment Specification " +
                    "at "+location+"."+key+ ".\"git\" has only 1 of \"username\" \"password\", " +
                    "while both or none are required.");
        }
        if (username != null && privateKey != null) {
            throw new ExperimentSpecificationParseException("Experiment Specification " +
                    "at "+location+"."+key+ ".\"git\" has both username and password auth, and private SSH key auth.");
        }

        return new GitFileSource(gitUrl,
                dir, branch,
                username, password,
                privateKey,
                filename);
    }

    @Nonnull
    private String getFileSourceBasename(@Nonnull FileSource.SourceType sourceType,
                                         @Nonnull String value,
                                         @Nonnull String location,
                                         @Nonnull Map in) throws ExperimentSpecificationParseException {
        switch (sourceType) {
            case BUNDLED: return value;
            case META: return value;
            case DIRECT: {
                Object path = in.get("path");
                if (path != null && path instanceof String) {
                    return (String)path;
                } else {
                    return location + (location.contains("execute") ? ".sh" : ".txt");
                }
            }
            case DOWNLOAD:
                try {
                    return new URL(value).getFile();
                } catch (MalformedURLException e) {
                    throw new ExperimentSpecificationParseException("Invalid URL \""+value+"\"", e);
                }
            case GIT: throw new RuntimeException("not yet supported");
            case GITHUB: throw new RuntimeException("not yet supported");
            default:
                throw new ExperimentSpecificationParseException("Unsupported SourceType: "+sourceType.name());
        }
    }


    @Nonnull
    private static String ensureEndsWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String ensureStartsWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }


    private void checkAllowedKeysWithFileSource(@Nonnull String location,
                                                @Nonnull Map<String, Object> map,
                                                String... allowedKeys) throws ExperimentSpecificationParseException {
        List<String> allKeys = new ArrayList<>(Arrays.asList(allowedKeys));
        allKeys.addAll(Arrays.asList(FileSource.SourceType.values()).stream()
                .map(Object::toString)
                .map(String::toLowerCase)
                .collect(Collectors.toList()));
        checkAllowedKeys(location, map, allKeys.toArray(new String[0]));
    }

    private void checkAllowedKeys(@Nonnull String location,
                                  @Nonnull Map<String, Object> map,
                                  String... allowedKeys) throws ExperimentSpecificationParseException {
        Set<String> allowedKeysSet = new HashSet<String>(Arrays.asList(allowedKeys));
        Set<String> unknownKeys = new HashSet<String>(map.keySet());
        unknownKeys.removeAll(allowedKeysSet);
        if (!unknownKeys.isEmpty()) {
            if (unknownKeys.size() == 1) {
                throw new ExperimentSpecificationParseException(
                        "Unknown key " + location + "." + unknownKeys.iterator().next());
            } else {
                throw new ExperimentSpecificationParseException(
                        "Unknown keys at " + location + ": " +
                                unknownKeys.stream().collect(Collectors.joining(",")));
            }
        }
    }
}
