package be.iminds.ilabt.jfed.rspec.rspec_source;

import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.model.imutable_impl.ImmutableModelRspec;
import be.iminds.ilabt.jfed.rspec.parser.RspecCompare;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.parser.StaxRspecParser;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.rspec.util.ProgressHandler;
import be.iminds.ilabt.jfed.util.common.ThreadFactoryUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.stream.XMLStreamException;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

/**
 * RspecSource
 * <p>
 * Note: this is a somewhat immutable class
 * <p>
 * RspecSource takes care of converting and caching different StringRspecs and ModelRspecs.
 * You can create one by providing either a StringRspec (assumed ModelRspec.RequestRspecSpecialCases.NONE) or a ModelRspec.
 * Then you can request any type of special case StringRspec, or a ModelRspec for any type of factory.
 * The RspecSource will convert what it has to create the requested rspec if needed, and cache the result so these conversions need to be done as little as possible.
 * <p>
 * This class tries to minimize blocking on RSpec-parsing by performing this in a separate thread.
 * If parsing hasn't completed yet, and a ModelRspec is requested, the call will block on this request until the parsing has completed.
 */
public interface RspecSource {
    boolean isXmlBased();

    @Nullable String getRspecXmlString();

    @Nullable String getRspecXmlString(@Nonnull ModelRspec.RequestRspecSpecialCases rspecSpecialCase);

    /**
     * @return a modelRspec, created by any mutable RspecFactory
     */
    @Nullable ModelRspec getMutableModelRspec();

    /**
     * @return a modelRspec, created by any immutable RspecFactory
     */
    @Nullable ImmutableModelRspec getImmutableModelRspec();

    /**
     * @return a modelRspec, created by any mutable RspecFactory
     */
    @Nullable ModelRspec getMutableModelRspec(ProgressHandler... progressHandlers);

    /**
     * @return a modelRspec, created by any immutable RspecFactory
     */
    @Nullable ImmutableModelRspec getImmutableModelRspec(ProgressHandler... progressHandlers);

    boolean isModelRspecAvailable(@Nonnull ModelRspecType modelRspecType);

    @Nullable
    <M extends ModelRspec> M getModelRspec(@Nonnull ModelRspecType modelRspecType, ProgressHandler... progressHandlers);

    @Nullable
    RspecParseException getLastInvalidRspecException();

    @Nullable
    default StringRspec getStringRspec() {
        return getStringRspec(ModelRspec.RequestRspecSpecialCases.NONE);
    }

    @Nullable
    StringRspec getStringRspec(@Nonnull ModelRspec.RequestRspecSpecialCases rspecSpecialCase);


    @Nonnull List<GeniUrn> getAllComponentManagerUrns();

    List<BasicStringRspec.BasicNodeInfo> getBasicNodeInfo();

    boolean isLosingData(ModelRspecType modelRspecType);

    List<RspecCompare.Difference> getLosingDataDifferences(ModelRspecType modelRspecType);

    class LosingDataInfo {
        private final boolean losingData;
        private final List<RspecCompare.Difference> losingDataDifferences;


        public LosingDataInfo(boolean losingData, List<RspecCompare.Difference> losingDataDifferences) {
            this.losingData = losingData;
            this.losingDataDifferences = losingDataDifferences;
        }

        public boolean isLosingData() {
            return losingData;
        }

        public List<RspecCompare.Difference> getLosingDataDifferences() {
            return losingDataDifferences;
        }
    }

}
